# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Docker shell helper class module.
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division

from pexpect import spawn


class DockerShell(object):
    """
    Shell helper class.

    This class wrapps a ``docker exec`` call to a given shell command and
    manages a pexpect spawn object for it.

    It implementes the ``__call__`` method that allows to call the objects
    as it were the contained shell.
    """

    def __init__(
            self, container, shell, prompt,
            prefix=None, timeout=None, encoding='utf-8'):
        self._container = container
        self._shell = shell
        self._prompt = prompt
        self._prefix = prefix
        self._timeout = timeout or -1
        self._encoding = encoding
        self._spawn = None

    def __call__(self, command):

        # Lazy-spawn
        if self._spawn is None:
            self._spawn = spawn(
                'docker exec -i -t {} {}'.format(
                    self._container, self._shell
                ),
                echo=False
            )
            # Cut output at first prompt
            self._spawn.expect(self._prompt, timeout=self._timeout)

        # Prefix command if required
        if self._prefix is not None:
            command = self._prefix + command

        self._spawn.sendline(command)
        self._spawn.expect(self._prompt, timeout=self._timeout)

        # Convert binary representation to unicode using encoding
        raw = self._spawn.before.decode(self._encoding)

        # Remove leading and trailing whitespaces and normalize newlines
        lines = raw.strip().replace('\r', '').splitlines()
        del raw

        # Remove echo command if it exists
        if len(lines) > 0 and lines[0] == command:
            lines.pop(0)

        return '\n'.join(lines)


__all__ = ['DockerShell']



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
topology_docker nodes manager module.

This module finds out which docker nodes plugins are installed and returns them
in a dictionary.
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division

import logging
from inspect import isclass
from traceback import format_exc
from collections import OrderedDict

from pkg_resources import iter_entry_points

from ..node import DockerNode


log = logging.getLogger(__name__)


def nodes(cache=True):
    """
    List all available nodes types.

    This function lists all available node types by discovering installed
    plugins registered in the entry point. This can be costly or error prone if
    a plugin misbehave. Because of this a cache is stored after the first call.

    :param bool cache: If ``True`` return the cached result. If ``False`` force
     reload of all plugins registered for the entry point.
    :rtype: dict
    :return: A dictionary associating the name of the node type and the class
     (subclass of :class:`topology_docker.platform.DockerNode`)
     implementing it:

     ::

        {
            'host': DockerNode,
            'openswitch': OpenSwitchNode
        }
    """

    # Return cached value if call is repeated
    if cache and hasattr(nodes, 'available'):
        return nodes.available

    # Add built-in node types
    available = OrderedDict()

    # Iterate over entry points
    for ep in iter_entry_points(group='topology_docker_node_10'):

        name = ep.name

        try:
            node = ep.load()
        except:
            log.error(
                'Unable to load node from plugin {}'.format(name)
            )
            log.debug(format_exc())
            continue

        if not isclass(node) or not issubclass(node, DockerNode):
            log.error(
                'Ignoring node "{}" as it doesn\'t '
                'match the required interface: '
                'Node not a subclass of DockerNode.'.format(name)
            )
            continue

        available[name] = node

    nodes.available = available
    return available


__all__ = ['nodes']



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Custom P4 Switch Topology Docker Node.

    http://p4.org/p4/an-open-source-p4-switch-with-sai-support/
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division

from subprocess import Popen
from shlex import split as shsplit

from topology_docker.node import DockerNode
from topology_docker.shell import DockerShell
from topology_docker.utils import ensure_dir


class P4SwitchNode(DockerNode):
    """
    Custom Barefoot P4 node for the Topology Docker platform engine.

    This custom node loads a P4 switch docker image built from the
    p4factory repository. The switch is configurable through switch-specific
    interfaces and the standard SAI interface (incomplete).

    See :class:`topology_docker.node.DockerNode`.
    """

    def __init__(
            self, identifier,
            image='topology/p4switch:latest',
            **kwargs):

        # Determine shared directory
        shared_dir = '/tmp/topology_{}_{}'.format(identifier, str(id(self)))
        ensure_dir(shared_dir)

        # Add binded directories
        binds = ['{}:/tmp'.format(shared_dir)]

        super(P4SwitchNode, self).__init__(
            identifier, image=image, binds=binds, **kwargs
        )

        # Behavioral model daemon process
        self._bm_daemon = None

        # Save location of the shared dir in host
        self.shared_dir = shared_dir

        # Add bash shell
        self._shells['bash'] = DockerShell(
            self.container_id, 'sh -c "TERM=dumb bash"', 'root@.*:.*# '
        )

        # Add SAI shell (https://github.com/p4lang/switchsai)
        self._shells['sai'] = DockerShell(
            self.container_id,
            'sh -c "TERM=dumb bash"',
            'root@.*:.*# ',
            '/p4factory/targets/switch/tests/pd_thrift/switch_sai_rpc-remote '
            '-h localhost:9092 '
        )

        # Add switchapi shell (https://github.com/p4lang/switchapi)
        self._shells['api'] = DockerShell(
            self.container_id,
            'sh -c "TERM=dumb bash"',
            'root@.*:.*# ',
            '/p4factory/targets/switch/tests/pd_thrift/switch_api_rpc-remote '
            '-h localhost:9091 '
        )

        # Add PD shells
        self._shells['pd_conn_mgr'] = DockerShell(
            self.container_id,
            'sh -c "TERM=dumb bash"',
            'root@.*:.*# ',
            '/p4factory/targets/switch/tests/pd_thrift/conn_mgr-remote '
            '-h localhost:9090 '
        )
        self._shells['pd_mc'] = DockerShell(
            self.container_id,
            'sh -c "TERM=dumb bash"',
            'root@.*:.*# ',
            '/p4factory/targets/switch/tests/pd_thrift/mc-remote '
            '-h localhost:9090 '
        )
        self._shells['pd_p4'] = DockerShell(
            self.container_id,
            'sh -c "TERM=dumb bash"',
            'root@.*:.*# ',
            '/p4factory/targets/switch/tests/pd_thrift/dc-remote '
            '-h localhost:9090 '
        )

    def notify_post_build(self):
        """
        Get notified that the post build stage of the topology build was
        reached.

        See :meth:`DockerNode.notify_post_build` for more information.
        """

        super(P4SwitchNode, self).notify_post_build()
        self._setup_system()

    def _setup_system(self):
        """
        Setup the P4 image for testing.

        #. Create the CPU virtual eth pair used by the P4 switch
        #. Bring up the interfaces required to run the switch
        #. Run the switch program
        """

        if self.metadata.get('autostart', True):

            self._docker_exec(
                'ip link add name veth250 type veth peer name veth251'
            )
            self._docker_exec('ip link set dev veth250 up')
            self._docker_exec('ip link set dev veth251 up')

            # Bring up all interfaces
            # FIXME: attach only interfaces brought up by the user
            for portlbl in self.ports:
                self.set_port_state(portlbl, True)

            args = [
                '/p4factory/targets/switch/behavioral-model',
                '--no-veth',
                '--no-pcap'
            ]

            # set openflow controller IP if necessary
            if self.metadata.get('ofip', None) is not None:
                args.extend(['--of-ip', self._ofip])

            # ifaces are ready in post_build
            # start the switch program with options:
            #  -i 1 -i 2 -i 3 -i 4 ...
            for iface in self.ports.values():
                args.extend(['-i', iface])

            # redirect stdout & stderr to log file
            # run in background
            args.extend(['&>/tmp/switch.log', '&'])

            # run command
            # TODO: check for posible error code
            self._bm_daemon = Popen(shsplit(
                'docker exec {} '.format(self.container_id) + ' '.join(args)
            ))


__all__ = ['P4SwitchNode']



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Custom Topology Docker Node for OpenvSwitch.

    http://openvswitch.org/

Check https://hub.docker.com/r/socketplane/openvswitch/ for docker container.
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division

from time import sleep
from subprocess import Popen
from shlex import split as shsplit

from topology_docker.node import DockerNode
from topology_docker.shell import DockerShell


class OpenvSwitchNode(DockerNode):
    """
    Custom Open vSwitch node for the Topology Docker platform engine.

    This custom node loads and starts an Open vSwitch image. It uses the OVS
    docker image. based on busybox OVS. It can be configured using
    ``ovs-vsctl``, ``ovs-ofctl`` and ``ovs-appclt`` through the default
    shell (sh). Interfaces are not auto-started by the node.

    See :class:`topology_docker.node.DockerNode`.
    """

    def __init__(
            self, identifier,
            image='topology/openvswitch:latest',
            **kwargs):

        # Supervisor daemon
        self._supervisord = None

        super(OpenvSwitchNode, self).__init__(
            identifier, image=image, command='sh', **kwargs
        )

        # Add bash shell
        self._shells['sh'] = DockerShell(
            self.container_id, 'sh', '/ .*#'
        )

    def notify_post_build(self):
        """
        Get notified that the post build stage of the topology build was
        reached.
        """
        super(OpenvSwitchNode, self).notify_post_build()

        # FIXME: this is a workaround
        self._docker_exec(
            "sed -i -e 's/port = 9001/port = 127.0.0.1:9001/g' "
            "-e 's/nodaemon=true/nodaemon=false/g' "
            "/etc/supervisord.conf"
        )

        self._supervisord = Popen(shsplit(
            'docker exec {} supervisord'.format(self.container_id)
        ))

        # Wait for the configure-ovs script to exit by polling supervisorctl
        config_timeout = 100
        i = 0
        while i < config_timeout:
            config_status = self._docker_exec(
                'supervisorctl status configure-ovs'
            )

            if 'EXITED' not in config_status:
                sleep(0.1)
            else:
                break
            i += 1

        if i == config_timeout:
            raise RuntimeError('configure-ovs did not exit!')


__all__ = ['OpenvSwitchNode']



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Custom Topology Docker Node for OpenSwitch.

    http://openswitch.net/
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division

from json import loads
from topology_docker.node import DockerNode
from topology_docker.shell import DockerShell
from topology_docker.utils import ensure_dir


SETUP_SCRIPT = """\
import logging
from sys import argv
from time import sleep
from os.path import exists
from json import dumps, loads
from shlex import split as shsplit
from subprocess import check_call, check_output
from socket import AF_UNIX, SOCK_STREAM, socket, gethostname

import yaml

config_timeout = 100
swns_netns = '/var/run/netns/swns'
hwdesc_dir = '/etc/openswitch/hwdesc'
db_sock = '/var/run/openvswitch/db.sock'
switchd_pid = '/var/run/openvswitch/ops-switchd.pid'
query = {
    'method': 'transact',
    'params': [
        'OpenSwitch',
        {
            'op': 'select',
            'table': 'System',
            'where': [],
            'columns': ['cur_hw']
        }
    ],
    'id': id(db_sock)
}
sock = None


def create_interfaces():
    # Read ports from hardware description
    with open('{}/ports.yaml'.format(hwdesc_dir), 'r') as fd:
        ports_hwdesc = yaml.load(fd)
    hwports = [str(p['name']) for p in ports_hwdesc['ports']]

    # Get list of already created ports
    not_in_swns = check_output(shsplit(
        'ls /sys/class/net/'
    )).split()
    in_swns = check_output(shsplit(
        'ip netns exec swns ls /sys/class/net/'
    )).split()

    create_cmd_tpl = 'ip tuntap add dev {hwport} mode tap'
    netns_cmd_tpl = 'ip link set {hwport} netns swns'
    rename_int = 'ip link set {portlbl} name {hwport}'

    # Save port mapping information
    mapping_ports = {}

    # Map the port with the labels
    for portlbl in not_in_swns:
        if portlbl in ['lo', 'oobm']:
            continue
        hwport = hwports.pop(0)
        mapping_ports[portlbl] = hwport
        logging.info(
            '  - Port {portlbl} moved to swns netns as {hwport}.'.format(
                **locals()
            )
        )
        check_call(shsplit(rename_int.format(**locals())))
        check_call(shsplit(netns_cmd_tpl.format(hwport=hwport)))

    # Writting mapping to file
    with open('/tmp/port_mapping.json', 'w') as json_file:
        json_file.write(dumps(mapping_ports))

    for hwport in hwports:
        if hwport in in_swns:
            logging.info('  - Port {} already present.'.format(hwport))
            continue

        logging.info('  - Port {} created.'.format(hwport))
        check_call(shsplit(create_cmd_tpl.format(hwport=hwport)))
        check_call(shsplit(netns_cmd_tpl.format(hwport=hwport)))

def cur_cfg_is_set():
    global sock
    if sock is None:
        sock = socket(AF_UNIX, SOCK_STREAM)
        sock.connect(db_sock)
    sock.send(dumps(query))
    response = loads(sock.recv(4096))
    return response['result'][0]['rows'][0]['cur_hw'] == 1


def main():

    if '-d' in argv:
        logging.basicConfig(level=logging.DEBUG)

    logging.info('Waiting for swns netns...')
    for i in range(0, config_timeout):
        if not exists(swns_netns):
            sleep(0.1)
        else:
            break
    else:
        raise Exception('Timed out while waiting for swns.')

    logging.info('Waiting for hwdesc directory...')
    for i in range(0, config_timeout):
        if not exists(hwdesc_dir):
            sleep(0.1)
        else:
            break
    else:
        raise Exception('Timed out while waiting for hwdesc directory.')

    logging.info('Creating interfaces...')
    create_interfaces()

    logging.info('Waiting for DB socket...')
    for i in range(0, config_timeout):
        if not exists(db_sock):
            sleep(0.1)
        else:
            break
    else:
        raise Exception('Timed out while waiting for DB socket.')

    logging.info('Waiting for cur_cfg...')
    for i in range(0, config_timeout):
        if not cur_cfg_is_set():
            sleep(0.1)
        else:
            break
    else:
        raise Exception('Timed out while waiting for cur_cfg.')

    logging.info('Waiting for switchd pid...')
    for i in range(0, config_timeout):
        if not exists(switchd_pid):
            sleep(0.1)
        else:
            break
    else:
        raise Exception('Timed out while waiting for switchd pid.')

    logging.info('Wait for final hostname...')
    for i in range(0, config_timeout):
        if gethostname() != 'switch':
            sleep(0.1)
        else:
            break
    else:
        raise Exception('Timed out while waiting for final hostname.')

if __name__ == '__main__':
    main()
"""


class OpenSwitchNode(DockerNode):
    """
    Custom OpenSwitch node for the Topology Docker platform engine.

    This custom node loads an OpenSwitch image and has vtysh as default
    shell (in addition to bash).

    See :class:`topology_docker.node.DockerNode`.
    """

    def __init__(
            self, identifier,
            image='topology/ops:latest',
            **kwargs):

        # Determine shared directory
        shared_dir = '/tmp/topology_{}_{}'.format(identifier, str(id(self)))
        ensure_dir(shared_dir)

        # Add binded directories
        binds = [
            '{}:/tmp'.format(shared_dir),
            '/dev/log:/dev/log',
            '/sys/fs/cgroup:/sys/fs/cgroup:ro'
        ]

        super(OpenSwitchNode, self).__init__(
            identifier, image=image, command='/sbin/init', binds=binds,
            **kwargs
        )

        # Save location of the shared dir in host
        self.shared_dir = shared_dir

        # Add vtysh (default) and bash shell
        self._shells['vtysh'] = DockerShell(
            self.container_id, 'vtysh', 'switch.*#'
        )
        self._shells['bash'] = DockerShell(
            self.container_id, 'sh -c "TERM=dumb bash"', 'bash-.*#'
        )
        self._shells['bash_swns'] = DockerShell(
            self.container_id,
            'sh -c "TERM=dumb ip netns exec swns bash"',
            'bash-.*#'
        )
        self._shells['vsctl'] = DockerShell(
            self.container_id, 'sh -c "TERM=dumb bash"', 'bash-.*#',
            prefix='ovs-vsctl ', timeout=60
        )

    def notify_post_build(self):
        """
        Get notified that the post build stage of the topology build was
        reached.

        See :meth:`DockerNode.notify_post_build` for more information.
        """
        super(OpenSwitchNode, self).notify_post_build()
        self._setup_system()

    def _setup_system(self):
        """
        Setup the OpenSwitch image for testing.

        #. Wait for daemons to converge.
        #. Assign an interface to each port label.
        #. Create remaining interfaces.
        """
        # Write and execute setup script
        setup_script = '{}/openswitch_setup.py'.format(self.shared_dir)
        with open(setup_script, 'w') as fd:
            fd.write(SETUP_SCRIPT)

        self._docker_exec('python /tmp/openswitch_setup.py')

        # Read back port mapping
        port_mapping = '{}/port_mapping.json'.format(self.shared_dir)
        with open(port_mapping, 'r') as fd:
            mappings = loads(fd.read())

        if hasattr(self, 'ports'):
            self.ports.update(mappings)
            return
        self.ports = mappings

    def set_port_state(self, portlbl, state):
        """
        Set the given port label to the given state.

        See :meth:`DockerNode.set_port_state` for more information.
        """
        iface = self.ports[portlbl]
        state = 'up' if state else 'down'

        not_in_netns = self._docker_exec('ls /sys/class/net/').split()
        prefix = '' if iface in not_in_netns else 'ip netns exec swns'

        command = '{prefix} ip link set dev {iface} {state}'.format(**locals())
        self._docker_exec(command)


__all__ = ['OpenSwitchNode']



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Custom Topology Docker Node for Ryu SND controller.

    http://osrg.github.io/ryu/
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division

from os import path
from time import sleep
from shutil import copy
from subprocess import Popen
from shlex import split as shsplit

from topology_docker.node import DockerNode
from topology_docker.shell import DockerShell
from topology_docker.utils import ensure_dir


class RyuControllerNode(DockerNode):
    """
    Custom Ryu SDN controller node capable of running custom apps.

    This custom node loads a Ryu docker image. It takes and runs a custom
    app by using ryu-manager.

    The default image is osrg/ryu from:

    - https://hub.docker.com/r/osrg/ryu/
    - https://github.com/osrg/dockerfiles/blob/master/ryu/Dockerfile

    See :class:`topology_docker.node.DockerNode`.
    """

    def __init__(
            self, identifier,
            image='topology/ryu:latest',
            **kwargs):

        # Determine shared directory
        shared_dir = '/tmp/topology_{}_{}'.format(identifier, str(id(self)))
        ensure_dir(shared_dir)

        # Add binded directories
        binds = ['{}:/tmp'.format(shared_dir)]

        super(RyuControllerNode, self).__init__(
            identifier, image=image, binds=binds, **kwargs
        )

        # Supervisor daemon
        self._supervisord = None

        # Save location of the shared dir in host
        self.shared_dir = shared_dir

        # Copy the ryu app into the container
        self.app_name = self.metadata.get('app', None)
        if self.app_name is not None:
            copy(self.app_name, self.shared_dir)

        # Add bash shell
        self._shells['bash'] = DockerShell(
            self.container_id, 'sh -c "TERM=dumb bash"', 'root@.*:.*# '
        )

    def notify_post_build(self):
        """
        Get notified that the post build stage of the topology build was
        reached.

        See :meth:`DockerNode.notify_post_build` for more information.
        """

        super(RyuControllerNode, self).notify_post_build()
        self._setup_system()

    def _setup_system(self):
        """
        Setup the controller image for testing.

        #. Bring up the ports connecting to the datapaths
        #. Run ryu-manager
        """

        # Ryu should be started by Topology
        if self.metadata.get('autostart', True):

            # Get the app file (or default)
            if self.app_name is not None:
                app_path = '/tmp/' + path.basename(self.app_name)
            else:
                app_path = '/root/ryu-master/ryu/app/simple_switch.py'

            # run ryu app using ryu-manager
            self._supervisord = Popen(shsplit(
                'docker exec {} '
                'sh -c "RYU_COMMAND=\'/root/ryu-master/bin/ryu-manager {} '
                '--verbose\' supervisord"'.format(
                    self.container_id,
                    app_path
                )
            ))

            # Wait for ryu-manager to start
            config_timeout = 100
            i = 0
            while i < config_timeout:
                config_status = self._docker_exec(
                    'supervisorctl status ryu-manager'
                )

                if 'RUNNING' not in config_status:
                    sleep(0.1)
                else:
                    break
                i += 1

            if i == config_timeout:
                raise RuntimeError(
                    'ryu-manager did not reach RUNNING state on supervisor!'
                )


__all__ = ['RyuControllerNode']
